from typing import Generic, TypeVar
from pydantic import BaseModel

from .fixed_basemodel import FixedBaseModel


R = TypeVar("R", FixedBaseModel, BaseModel)


class IPaginatedResponse(FixedBaseModel, Generic[R]):
    page_size: int
    page_number: int
    items: list[R]
    total_pages: int
    total_count: int


from uuid import UUID
from attp_client.misc.fixed_basemodel import FixedBaseModel


class ITool(FixedBaseModel):
    """Tool interface definition"""
    id: UUID | str
    name: str
    description: str | None = None
    schema_id: str | None = None
    schema: dict | None = None
    return_direct: bool = False
    schema_ver: str = "1.0"
    timeout_ms: float = 20000
    idempotent: bool = False
    config: dict | None = None
    
    catalog: str


from uuid import UUID
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.types.chatmode import ChatModeEnum


class IChatDTO(FixedBaseModel):
    name: str
    mode: ChatModeEnum
    platform: str
    agent_name: str | None = None
    agent_id: int | None = None
    responsible: int | None = None
    client_id: str | None = None
    

class IChatResponse(IChatDTO):
    id: UUID


import asyncio
from contextvars import ContextVar
from typing import Any, AsyncIterable, Callable, TypeVar, overload
import msgpack
from pydantic import TypeAdapter
from reactivex import Subject, defer, empty, from_future, of, operators as ops, throw, timer
from reactivex.scheduler.eventloop import AsyncIOScheduler
from attp_core.rs_api import PyAttpMessage, AttpCommand

from attp_client.errors.correlated_rpc_exception import CorrelatedRPCException
from attp_client.errors.dead_session import DeadSessionError
from attp_client.errors.serialization_error import SerializationError
from attp_client.interfaces.error import IErr
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable
from attp_client.session import SessionDriver
from attp_client.utils.context_awaiter import ContextAwaiter
from attp_client.utils.stream_receiver import StreamReceiver


T = TypeVar("T")
S = TypeVar("S")


class AttpRouter:
    def __init__(
        self, 
        responder: Subject[PyAttpMessage],
        session: SessionDriver
    ) -> None:
        self.responder = responder
        self.session = session
        self.context = ContextVar[str | None]("session_context", default=None)
    
    def convert_message(self, expected_type: type[T], message: PyAttpMessage) -> T | Any:
        response = self.__format_response(expected_type=expected_type, response_data=message)
        return response
    
    @overload
    async def send(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 20,
    ) -> Any: ...
    
    @overload
    async def send(
        self, 
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 20, *,
        expected_response: type[T],
    ) -> T | Any: ...
    
    async def send(
        self, 
        route: str, 
        data: FixedBaseModel | Serializable | None = None,
        timeout: float = 50, *,
        expected_response: type[T] | None = None
    ) -> T | Any:
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)
        # correlation_id = await self.session.send_message(pattern, data)
        
        responder = ContextAwaiter[Any](defer(
            lambda _: (
                from_future(asyncio.ensure_future(self.session.send_message(route=route, data=data))).pipe(
                    ops.flat_map(
                        lambda cid: empty().pipe(
                            ops.concat(self.__pipe_filter(cid, timeout=timeout)),
                        )
                    )
                )
            )
        ))
        
        response_data = await responder.wait()
        
        return self.__format_response(expected_type=expected_response or Any, response_data=response_data)
    
    @overload
    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 50,
        *,
        formatter: Callable[[PyAttpMessage], S | None],
    ) -> AsyncIterable[S]: ...

    @overload
    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 50,
        *,
        formatter: None = ...,
    ) -> AsyncIterable[Any]: ...

    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = None,
        timeout: float = 50,
        *,
        formatter: Callable[[PyAttpMessage], S | None] | None = None,
    ) -> AsyncIterable[Any] | AsyncIterable[S]:
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)
        
        stream = StreamReceiver(
            from_future(asyncio.ensure_future(self.session.send_message(route=route, data=data))).pipe(
                ops.flat_map(
                    lambda cid: empty().pipe(
                        ops.concat(self.__stream_pipe_filter(cid, timeout=timeout)),
                    )
                )
            ),
            formatter=formatter,
        )
        
        return stream
        
    
    async def emit(self, route: str, data: FixedBaseModel | Serializable | None = None):
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)
        
        await self.session.emit_message(route, data)
    
    def __pipe_filter(self, awaiting_correlation_id: bytes, timeout: float):
        loop = asyncio.get_event_loop()
        asyncio_scheduler = AsyncIOScheduler(loop)
        # print(awaiting_correlation_id)
        return self.responder.pipe(
            ops.subscribe_on(asyncio_scheduler),
            ops.filter(lambda pair: pair.correlation_id == awaiting_correlation_id),
            ops.flat_map(lambda r: throw(CorrelatedRPCException.from_err_object(correlation_id=r.correlation_id or b'<nocorrid>', err=IErr.mps(r.payload) if r.payload else IErr(detail={"code": "ErrorWithoutPayload"}))) if r.command_type == AttpCommand.ERR else of(r)),
            #######################################
            ##     This is RPC Defer Handler     ##
            #######################################
            ops.timeout_with_mapper(
                timer(timeout, scheduler=asyncio_scheduler),
                lambda i: (
                    timer(timeout, scheduler=asyncio_scheduler) if getattr(i, "frame_type", None) == AttpCommand.DEFER else of(None)
                ),
                throw(TimeoutError("ATTP response failed."))
            ),
            ops.filter(lambda pair: pair.command_type == AttpCommand.ACK),
            ops.first(),
        )

    def __stream_pipe_filter(self, awaiting_correlation_id: bytes, timeout: float):
        loop = asyncio.get_event_loop()
        asyncio_scheduler = AsyncIOScheduler(loop)
        return self.responder.pipe(
            ops.subscribe_on(asyncio_scheduler),
            ops.filter(lambda pair: pair.correlation_id == awaiting_correlation_id),
            ops.flat_map(lambda r: throw(CorrelatedRPCException.from_err_object(correlation_id=r.correlation_id or b'<nocorrid>', err=IErr.mps(r.payload) if r.payload else IErr(detail={"code": "ErrorWithoutPayload"}))) if r.command_type == AttpCommand.ERR else of(r)),
            #######################################
            ##     This is RPC Defer Handler     ##
            #######################################
            # ops.timeout_with_mapper(
            #     timer(timeout, scheduler=asyncio_scheduler),
            #     lambda i: (
            #         timer(timeout, scheduler=asyncio_scheduler) if getattr(i, "frame_type", None) == AttpCommand.DEFER else of(None)
            #     ),
            #     throw(TimeoutError("ATTP stream failed."))
            # ),
            ops.filter(lambda pair: pair.command_type in (
                AttpCommand.STREAMBOS,
                AttpCommand.CHUNK,
                AttpCommand.STREAMEOS,
            )),
            ops.take_while(lambda pair: pair.command_type != AttpCommand.STREAMEOS, inclusive=True),
        )

    def __format_response(self, expected_type: Any, response_data: PyAttpMessage):
        if issubclass(expected_type, FixedBaseModel):
            if not response_data.payload:
                raise SerializationError(f"Nonetype payload received from session while expected type {expected_type.__name__}")
            try:
                return expected_type.mps(response_data.payload)
            except Exception as e:
                raise SerializationError(str(e))
        
        serialized = msgpack.unpackb(response_data.payload) if response_data.payload else None
        
        if expected_type is not None:
            return serialized
        
        return TypeAdapter(expected_type, config={"arbitrary_types_allowed": True}).validate_python(serialized)



from typing import overload
from uuid import UUID
from attp_client.interfaces.objects.chat import IChatDTO, IChatResponse
from attp_client.misc.paginated import IPaginatedResponse
from attp_client.misc.serializable import Serializable
from attp_client.router import AttpRouter
# from attp_client.session import SessionDriver
from attp_client.types.chatmode import ChatModeEnum


class AttpChats:
    def __init__(self, router: AttpRouter) -> None:
        self.router = router

    @overload
    async def create_chat(
        self,
        name: str,
        platform: str,
        mode: ChatModeEnum = ...,
    ): ...
    
    @overload
    async def create_chat(
        self,
        name: str,
        platform: str,
        mode: ChatModeEnum = ...,
        *,
        agent_name: str,
    ): ...
    
    @overload
    async def create_chat(
        self,
        name: str,
        platform: str,
        mode: ChatModeEnum = ...,
        *,
        agent_id: int,
    ): ...
    
    @overload
    async def create_chat(
        self,
        name: str,
        platform: str,
        mode: ChatModeEnum = ...,
        *,
        responsible: int,
        client_id: str | None = None,
    ): ...

    async def create_chat(
        self, 
        name: str, 
        platform: str,
        mode: ChatModeEnum = ChatModeEnum.AGENT_AUTOPILOT,
        *,
        agent_name: str | None = None,
        agent_id: int | None = None,
        responsible: int | None = None,
        client_id: str | None = None,
    ):
        if agent_name and agent_id:
            raise ValueError("You can only provide either 'agent_name' or 'agent_id', not both.")

        if not agent_name and not agent_id:
            raise ValueError("You must provide either 'agent_name' or 'agent_id'.")

        payload = IChatDTO(
            name=name,
            mode=mode,
            platform=platform,
            agent_name=agent_name,
            agent_id=agent_id,
            responsible=responsible,
            client_id=client_id,
        )

        response = await self.router.send("chats:create", payload, expected_response=IChatResponse)
        
        return response
    
    async def get_chats(self):
        response = await self.router.send("chats:list", expected_response=IPaginatedResponse[IChatResponse])
        
        return response
    
    async def get_chat(self, chat_id: UUID):
        response = await self.router.send("chats:specific", Serializable[dict[str, UUID]]({"chat_id": chat_id}), expected_response=IChatResponse)
        
        return response
    
    async def delete_chat(self, chat_id: UUID):
        await self.router.emit("chats:delete", Serializable[dict[str, UUID]]({"chat_id": chat_id}))

    async def change_agent(self, chat_id: UUID, new_agent_id: int):
        await self.router.emit("chats:agent", Serializable[dict[str, UUID | int]]({"chat_id": chat_id, "new_agent_id": new_agent_id}))
    
    


class DeadSessionError(Exception):
    def __init__(self, organization_id: int) -> None:
        super().__init__()
        self.organization_id = organization_id
    
    def __str__(self) -> str:
        return f"Cannot perform any actions over dead ATTP session (organization_id={self.organization_id})"


import asyncio
from functools import cached_property
from logging import Logger, getLogger
from typing import Any, Callable
from attp_core.rs_api import AttpClientSession, Limits
from reactivex import Subject, operators as ops
from reactivex.scheduler.eventloop import AsyncIOScheduler
from attp_core.rs_api import PyAttpMessage

from attp_client.catalog import AttpCatalog
from attp_client.errors.dead_session import DeadSessionError
from attp_client.inference import AttpInferenceAPI
from attp_client.interfaces.catalogs.catalog import ICatalogResponse
from attp_client.interfaces.error import IErr
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable
from attp_client.objects.agents import AttpAgents
from attp_client.objects.chats import AttpChats
from attp_client.router import AttpRouter
from attp_client.session import SessionDriver
from attp_client.tools import ToolsManager
from attp_client.types.route_mapping import AttpRouteMapping, RouteType
from attp_client.utils import envelopizer

from attp_core.rs_api import AttpCommand, init_logging

from attp_client.utils.stream_object import StreamObject
from attp_client.utils.trigger_callable import trigger_callable


class ATTPClient:
    
    is_connected: bool
    client: AttpClientSession
    session: SessionDriver | None
    routes: list[AttpRouteMapping]
    inference: AttpInferenceAPI
    catalogs: list[AttpCatalog]

    _tools: ToolsManager | None

    def __init__(
        self,
        agt_token: str,
        organization_id: int,
        *,
        connection_url: str | None = None,
        reconnect: bool = False,
        max_retries: int = 20,
        limits: Limits | None = None,
        logger: Logger | None = None,
        verbose: bool = False
    ):
        self.__agt_token = agt_token
        self.organization_id = organization_id
        self.connection_url = connection_url or "attp://localhost:6563"
        
        self.session = None
        self.max_retries = max_retries
        self.limits = limits or Limits(max_payload_size=50000)
        self.client = AttpClientSession(self.connection_url, limits=self.limits)
        self.logger = logger or getLogger("Ascender Framework")
        self.reconnect = reconnect
        
        self.route_increment_index = 2
        
        self.responder = Subject[PyAttpMessage]()
        self.routes = []
        self.catalogs = []
        self.disposable = None
        self._client = None
        self.verbose = verbose
        
        self._tools = None
        
        if self.verbose:
            init_logging()
    
    async def connect(self):
        # Open the connection
        client = await self.client.connect(self.max_retries)
        self._client = client
        
        if not client.session:
            raise ConnectionError("Failed to connect to ATTP server after 10 attempts!")
        
        self.session = SessionDriver(
            client.session, 
            agt_token=self.__agt_token, 
            organization_id=self.organization_id,
            # route_mappings=self.routes,
            factory=self.client,
            on_reconnect=self._reconnect if self.reconnect else None,
            logger=self.logger or getLogger("Ascender Framework")
        )
        asyncio.create_task(self.session.start_listener())
        # Send an authentication frame as soon as connection estabilishes with agenthub
        self.add_event_handler("tools:call", "message", self._tool_callback)
        self.add_event_handler("catalogs:tools:list", "message", self._list_tools)
        
        await self.session.authenticate(self.routes)
        asyncio.create_task(self.session.listen(self.responder))
        
        self.router = AttpRouter(self.responder, self.session)
        self.inference = AttpInferenceAPI(self.router)
        self.chats = AttpChats(self.router)
        self.agents = AttpAgents(self.router)
        
        self.disposable = self.responder.pipe(
            ops.subscribe_on(AsyncIOScheduler(asyncio.get_event_loop())),
        ).subscribe(
            on_next=lambda item: trigger_callable(self._handle_incoming, (item,)),
            on_error=lambda e: self.logger.error(f"Error in responder stream: {e}"),
        )
    
    async def _reconnect(self):
        self.logger.info("Attempting to reconnect to ATTP server...")
        await self.connect()

    async def close(self):
        if self.session:
            self.session.on_reconnect = None
            
            if self.disposable:
                self.disposable.dispose()
            
            await self.session.close()
            
            self.session = None
            self.is_connected = False

    @property
    def tools(self):
        if not self._tools:
            self._tools = ToolsManager(self.router)
        
        return self._tools
    
    async def catalog(self, catalog_name: str):
        if any(c.catalog_name == catalog_name for c in self.catalogs):
            return next(c for c in self.catalogs if c.catalog_name == catalog_name)
        
        catalog = await self.router.send(
            "tools:catalogs:specific", 
            Serializable[dict[str, str]]({"catalog_name": catalog_name}),
            timeout=10,
            expected_response=ICatalogResponse
        )
        self.catalogs.append(
            AttpCatalog(id=catalog.catalog_id, catalog_name=catalog_name, manager=self.tools)
        )

        return self.catalogs[-1] # Return the newly added catalog

    async def close_catalog(self, catalog: AttpCatalog):
        catalog.detach_all_tools()
        self.catalogs.remove(catalog)

    async def _tool_callback(self, message: PyAttpMessage):
        if not self.session:
            raise DeadSessionError(self.organization_id)
        
        if not message.correlation_id:
            await self.session.send_error(IErr(
                detail={"message": "Correlation ID was missing in the message.", "code": "MissingCorrelationId"},
            ), route=message.route_id)
            return
        
        print("TOOL CALLBACK MESSAGE:", message.payload)
        try:
            envelope = envelopizer.envelopize(message)
        except ValueError as e:
            await self.session.send_error(IErr(
                detail={"message": str(e), "code": "InvalidPayload"},
            ), correlation_id=message.correlation_id, route=message.route_id)
            return

        catalog = next((c for c in self.catalogs if c.catalog_name == envelope.catalog), None)
        if not catalog:
            await self.session.send_error(IErr(
                detail={"message": f"Catalog with name {envelope.catalog} not found.", "code": "NotFoundError"},
            ), route=message.route_id, correlation_id=message.correlation_id)
            return

        
        response = await catalog.handle_callback(envelope)
        
        if isinstance(response, IErr):
            await self.session.send_error(response, route=message.route_id, correlation_id=message.correlation_id)
            return

        if not isinstance(response, FixedBaseModel) and not isinstance(response, Serializable):
            response = Serializable[Any](response)

        await self.session.respond(route=message.route_id, correlation_id=message.correlation_id, payload=response)

    async def _list_tools(self, message: PyAttpMessage):
        if not self.session:
            raise DeadSessionError(self.organization_id)
        
        if not message.payload:
            await self.session.send_error(IErr(
                detail={"message": "Payload was missing in the message.", "code": "MissingPayload"},
            ), route=message.route_id)
            return
        
        deserialized = Serializable[dict[str, Any]].mps(message.payload)
        
        if deserialized.data.get("catalog_name") is None:
            await self.session.send_error(IErr(
                detail={"message": "Catalog name was missing in the payload.", "code": "MissingCatalogName"},
            ), route=message.route_id, correlation_id=message.correlation_id)
            return
        catalog_name = deserialized.data["catalog_name"]
        tools = self.tools.get_tools(catalog_name=catalog_name)
        
        return Serializable[dict[str, Any]]({
            "tools": [tool.model_dump(mode="json") for tool in tools]
        })
    
    async def _handle_incoming(self, message: PyAttpMessage):
        if message.route_id <= 1:
            return
        
        relevant_route = next((route for route in self.routes if route.route_id == message.route_id), None)
        
        if not relevant_route:
            if self.logger:
                self.logger.warning(f"Received message for unknown route ID {message.route_id}. Ignoring.")
            return
        
        response = await relevant_route.callback(message)
        
        if message.command_type == AttpCommand.CALL:
            if not self.session:
                raise DeadSessionError(self.organization_id)
            
            if not message.correlation_id:
                await self.session.send_error(IErr(
                    detail={"message": "Correlation ID was missing in the message.", "code": "MissingCorrelationId"},
                ), route=message.route_id)
                return
            if isinstance(response, StreamObject):
                iterator = response.iterate()
                if not iterator:
                    return

                # Signal beginning of stream
                await self.session.respond_stream(
                    route=message.route_id,
                    correlation_id=message.correlation_id,
                    command_type=AttpCommand.STREAMBOS,
                )

                if response.is_async:
                    async for chunk in iterator:  # type: ignore
                        await self.session.respond_stream(
                            route=message.route_id,
                            correlation_id=message.correlation_id,
                            payload=chunk,
                            command_type=AttpCommand.CHUNK,
                        )
                else:
                    for chunk in iterator:  # type: ignore
                        await self.session.respond_stream(
                            route=message.route_id,
                            correlation_id=message.correlation_id,
                            payload=chunk,
                            command_type=AttpCommand.CHUNK,
                        )

                # End of stream
                await self.session.respond_stream(
                    route=message.route_id,
                    correlation_id=message.correlation_id,
                    command_type=AttpCommand.STREAMEOS,
                )
                return

            await self.session.respond(
                route=message.route_id,
                correlation_id=message.correlation_id,
                payload=response
            )
    
    def add_event_handler(
        self, 
        pattern: str, 
        route_type: RouteType,
        callback: Callable[..., Any],
    ):
        if route_type in ["connect", "disconnect"]:
            self.routes.append(
                AttpRouteMapping(
                    pattern=pattern,
                    route_id=0,
                    route_type=route_type,
                    callback=callback
                )
            )
            return
        
        self.routes.append(
            AttpRouteMapping(
                pattern=pattern, 
                route_id=self.route_increment_index, 
                route_type=route_type, 
                callback=callback
            )
        )
        
        self.route_increment_index += 1


from enum import Enum


class MessageDataTypeEnum(Enum):
    """
    Enumeration for data types available to be sent by users/AI
    """

    JPEG_IMAGE = 'jpeg_image'  # .jpeg .jpg
    FFMPEG_VIDEO = 'ffmpeg_video'  # mp4
    PLAIN_TEXT = 'plain_text'  # txt, that would be also a simple chat message
    WAVE_AUDIO = 'wave_audio'  # wav
    PLACEHOLDER_KEY = 'placeholder_key'
    NONE = 'none'



from typing import Annotated
from pydantic import Field
from typing_extensions import Doc

from attp_client.interfaces.route_mappings import IRouteMapping
from attp_client.misc.fixed_basemodel import FixedBaseModel


class IHello(FixedBaseModel):
    proto: Annotated[str, Doc("Protocol ID")] = Field("ATTP")
    ver: Annotated[str, Doc("Semver for example: 1.0")] = Field("1.0")
    caps: Annotated[list[str], Doc("Capability flags e.g. ['schemas/json', 'stream', 'hb']")] = Field(default_factory=list)
    mapping: Annotated[list[IRouteMapping], Doc("Router pattern mappings, ATTP for optimization uses numbers and IDs for routing requests instead of human friendly patterns")] 


from enum import Enum


class MessageTypeEnum(Enum):
    TOOL_STATE_EVENT = 'tool_state_event'

    OPERATOR_CALLED_EVENT = 'operator_called_event'

    SYSTEM_EVENT = 'system_event'  # user is created, or added to chat or smth
    SYSTEM_MESSAGE = 'system_message'

    USER_MESSAGE = 'user_message'
    AI_MESSAGE = 'ai_message'
    #
    OPERATOR_MESSAGE = 'operator_message'
    CUSTOMER_MESSAGE = 'customer_message'

    PLACEHOLDER_MESSAGE = 'placeholder_message'  # content is the key for placeholder



from typing import Sequence

from attp_client.errors.not_found import NotFoundError
from attp_client.interfaces.route_mappings import IRouteMapping

from attp_client.types.route_mapping import RouteType


def resolve_route_by_id(route_type: RouteType, pattern: str, route_mapping: Sequence[IRouteMapping]):
    route = next(
        (route for route in route_mapping if route.pattern == pattern and route.route_type == route_type),
        None
    )
    
    if not route:
        raise NotFoundError(f"Route {pattern} not found.")
    
    return route


import asyncio
from typing import Generic, TypeVar
from reactivex import Observable


T = TypeVar("T")


class ContextAwaiter(Generic[T]):
    def __init__(
        self, 
        observable: Observable[T],
    ) -> None:
        self.observable = observable
        self.response = asyncio.Future[T]()
        self.event = asyncio.Event()
        self.subscription = None
        
    async def wait(self) -> T:
        self.subscription = self.observable.subscribe(
            on_next=self.__define_response,
            on_error=self.__set_error
        )
        
        return await self.response
    
    def __define_response(self, resp: T):
        if self.subscription:
            self.subscription.dispose()
            self.subscription = None
        
        if not self.response.done():
            self.response.set_result(resp)
        self.event.set()
    
    def __set_error(self, exc: Exception):
        if not self.response.done():
            self.response.set_exception(exc)
        
        self.event.set()


from typing import Annotated, Sequence
from typing_extensions import Doc

from attp_client.interfaces.route_mappings import IRouteMapping
from attp_client.misc.fixed_basemodel import FixedBaseModel


class IReady(FixedBaseModel):
    session: Annotated[str, Doc("Server-issued session id (per connection)")] # REQUIRED
    server_time: Annotated[str | None, Doc("Server ISO time for skew hints")] = None
    server_routes: Sequence[IRouteMapping]


from inspect import isasyncgenfunction, isfunction
from typing import AsyncGenerator, AsyncIterable, AsyncIterator, Iterable, Iterator, TypeAlias

from attp_client.misc.fixed_basemodel import FixedBaseModel

from collections.abc import AsyncIterable as AsyncIterableABC, AsyncIterator as AsyncIteratorABC


IterateWrapper: TypeAlias = (Iterable[FixedBaseModel] | Iterator[FixedBaseModel] | 
                             AsyncGenerator[FixedBaseModel, None] | AsyncIterable[FixedBaseModel] |
                            AsyncIterator[FixedBaseModel]
                            )

class StreamObject:    
    def __init__(
        self, 
        _iterable: IterateWrapper | None = None
    ) -> None:
        self._iterable = _iterable
        self.is_async = False
        
        if isinstance(_iterable, (AsyncIterableABC, AsyncIteratorABC)):
            self.is_async = True
    
    def iterate(self) -> IterateWrapper | None:
        return self._iterable


import asyncio
from typing import Any, AsyncIterator, Callable, Generic, TypeVar

import msgpack
from reactivex import Observable
from attp_core.rs_api import PyAttpMessage


T = TypeVar("T")


class StreamReceiver(Generic[T]):
    def __init__(
        self,
        observable: Observable[PyAttpMessage],
        *,
        formatter: Callable[[PyAttpMessage], T | None] | None = None,
    ) -> None:
        self.observable = observable
        self.formatter = formatter or self.__default_formatter
        self.queue: asyncio.Queue[PyAttpMessage] = asyncio.Queue()
        self.completed = asyncio.Event()
        self.subscription = None
        self.error: Exception | None = None
        self.started = False

    def __aiter__(self) -> AsyncIterator[T]:
        return self.__iter_stream()

    def __subscribe(self) -> None:
        if self.started:
            return
        self.started = True
        self.subscription = self.observable.subscribe(
            on_next=self.queue.put_nowait,
            on_error=self.__on_error,
            on_completed=self.completed.set,
        )

    async def __iter_stream(self) -> AsyncIterator[T]:
        self.__subscribe()

        try:
            while True:
                if self.error:
                    raise self.error

                if self.completed.is_set() and self.queue.empty():
                    break

                message = await self.queue.get()
                formatted = self.formatter(message)
                if formatted is None:
                    continue
                yield formatted
        finally:
            self.__dispose()

    def __on_error(self, exc: Exception):
        self.error = exc
        self.completed.set()

    def __dispose(self):
        if self.subscription:
            self.subscription.dispose()
            self.subscription = None

    def __default_formatter(self, msg: PyAttpMessage) -> Any | None:
        if msg.payload is None:
            return None
        return msgpack.unpackb(msg.payload)



from typing import Any

import msgpack
from pydantic import TypeAdapter
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable


def serialize(data: bytes | None, model: type[FixedBaseModel] | Any) -> FixedBaseModel | None:
    if not data:
        return None
    
    if issubclass(model, FixedBaseModel):
        return model.mps(data)
    
    return TypeAdapter(model).validate_python(msgpack.unpackb(data))


def deserialize(data: FixedBaseModel | Serializable | Any | None) -> bytes | None:
    if data is None:
        return None

    if isinstance(data, (FixedBaseModel, Serializable)):
        return data.mpd()
    
    return msgpack.packb(data)


from collections import defaultdict
from typing import Any, Sequence
from uuid import UUID, uuid4
from attp_client.errors.not_found import NotFoundError
from attp_client.interfaces.catalogs.tools.tool import ITool
from attp_client.misc.serializable import Serializable
from attp_client.router import AttpRouter


class ToolsManager:
    tools: dict[str, list[ITool]] # catalog_name, tool
    
    def __init__(self, router: AttpRouter) -> None:
        self.router = router
        self.tools = defaultdict(list)
    
    def register(
        self, 
        catalog_name: str,
        name: str, 
        description: str | None = None,
        schema_id: str | None = None,
        schema: dict | None = None,
        *,
        return_direct: bool = False,
        schema_ver: str = "1.0",
        timeout_ms: float = 20000,
        idempotent: bool = False,
        configs: Any | None = None
    ) -> UUID:
        tool_id = uuid4()
        
        self.tools[catalog_name].append(ITool(
            id=tool_id,
            name=name,
            description=description,
            schema_id=schema_id,
            return_direct=return_direct,
            schema=schema,
            schema_ver=schema_ver,
            timeout_ms=timeout_ms,
            idempotent=idempotent,
            config=configs,
            catalog=catalog_name
        ))
        
        return tool_id
    
    def unregister(
        self,
        catalog_name: str,
        tool_id: UUID | Sequence[UUID]
    ) -> UUID | list[UUID]:
        if catalog_name not in self.tools:
            raise NotFoundError(f"Catalog {catalog_name} not found in the tools manager.")
        
        if isinstance(tool_id, UUID):
            if str(tool_id) in [str(t.id) for t in self.tools[catalog_name]]:
                self.tools[catalog_name] = [
                    t for t in self.tools[catalog_name] if str(t.id) != str(tool_id)
                ]
                return tool_id
            raise NotFoundError(f"Tool {tool_id} not found in the catalog {catalog_name}.")
        
        removed_ids = []
        for tid in tool_id:
            if str(tid) in [str(t.id) for t in self.tools[catalog_name]]:
                self.tools[catalog_name] = [
                    t for t in self.tools[catalog_name] if str(t.id) != str(tid)
                ]
                removed_ids.append(tid)
        
        return removed_ids
    
    def get_tools(self, catalog_name: str) -> list[ITool]:
        return self.tools.get(catalog_name, [])
    
    def get_tool(self, catalog_name: str, tool_id: UUID) -> ITool:
        # Use next(...) to find the tool with the matching id
        tool = next((t for t in self.tools.get(catalog_name, []) if str(t.id) == str(tool_id)), None)
        
        if tool:
            return tool
        
        raise NotFoundError(f"Tool {tool_id} not found in the catalog {catalog_name}.")


from attp_client.interfaces.catalogs.tools.envelope import IEnvelope
from attp_core.rs_api import PyAttpMessage


def envelopize(message: PyAttpMessage) -> IEnvelope:
    if not message.payload:
        raise ValueError("Message payload is empty, cannot envelopize.")

    return IEnvelope.mps(message.payload)


from typing import Any

from attp_client.interfaces.objects.agent import IAgentDTO, IAgentResponse
from attp_client.misc.serializable import Serializable
from attp_client.router import AttpRouter


class AttpAgents:
    def __init__(self, router: AttpRouter) -> None:
        self.router = router

    async def create_agent(self, data: IAgentDTO) -> IAgentResponse:
        """Create a new agent."""
        response = await self.router.send("agents:create", data, expected_response=IAgentResponse)
        return response

    async def get_agent(self, agent_id: int | None = None, agent_name: str | None = None) -> IAgentResponse:
        """Get a single agent by id or by name."""
        if not agent_id and not agent_name:
            raise ValueError("You must provide either 'agent_id' or 'agent_name'.")
        if agent_id and agent_name:
            raise ValueError("Provide only one of 'agent_id' or 'agent_name', not both.")

        payload: dict[str, Any] = {}
        if agent_id is not None:
            payload["agent_id"] = agent_id
        if agent_name is not None:
            payload["agent_name"] = agent_name

        response = await self.router.send(
            "agents:get",
            Serializable[dict[str, Any]](payload),
            expected_response=IAgentResponse,
        )
        return response

    async def update_agent(self, agent_id: int, data: IAgentDTO) -> IAgentResponse:
        """Update an existing agent by id."""
        payload = Serializable[dict[str, Any]]({
            "agent_id": agent_id,
            "data": data.model_dump(mode="json"),
        })
        response = await self.router.send(
            "agents:update",
            payload,
            expected_response=IAgentResponse,
        )
        return response

    async def delete_agent(self, agent_id: int) -> None:
        """Delete an agent by id."""
        await self.router.emit("agents:delete", Serializable[dict[str, int]]({"agent_id": agent_id}))



from typing import Any, Self

import msgpack
from pydantic import BaseModel


class FixedBaseModel(BaseModel):
    @classmethod
    def serialize(cls, entity, **kwargs):
        serialized_data = {
            key: value
            for key, value in entity.__dict__.items()
            if not callable(value) and not key.startswith('_')
        }

        for key, value in kwargs.items():
            serialized_data[key] = value

        # noinspection PyArgumentList
        pydantic_instance = cls(**serialized_data)

        return pydantic_instance

    @classmethod
    def s(cls, entity, **kwargs) -> Self:
        return cls.serialize(entity, **kwargs)

    @classmethod
    def mps(
        cls, 
        obj: bytes,
        strict: bool | None = None,
        from_attributes: bool | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
        mp_configs: dict[str, Any] | None = None
    ):
        """
        Message Pack Serialize
        
        Serializes and unpacks the model from the binary by utilizing Message Pack library.

        Opposite method: `mpd(...)`

        Parameters
        ----------
        obj : bytes
            Binary packed by Message Pack object.
        """
        obj = msgpack.unpackb(obj, **(mp_configs or {}))
        
        return cls.model_validate(obj, strict=strict, from_attributes=from_attributes, context=context, by_alias=by_alias, by_name=by_name)
    
    def mpd(self, mp_configs: dict[str, Any] | None = None, **kwargs) -> bytes | None:
        """
        Message Pack Dump
        
        Dumps and packs the model to the binary by utilizing Message Pack library.
        
        Opposite method: `mps(...)`
        """
        return msgpack.packb(self.model_dump(mode="json", **kwargs), **(mp_configs or {}))


from logging import Logger, getLogger
from typing import Any, AsyncIterable, Literal, Sequence, overload
from uuid import UUID
from attp_client.interfaces.inference.message import IMessageResponse, IMessageDTOV2
from attp_client.misc.serializable import Serializable
from attp_client.router import AttpRouter


class AttpInferenceAPI:
    """
    AttpInferenceAPI provides methods to interact with the inference API of the AgentHub.
    """
    def __init__(
        self,
        router: AttpRouter,
        logger: Logger = getLogger("Ascender Framework")
    ) -> None:
        self.router = router
        self.logger = logger
    
    async def create_chat(
        self,
        name: str,
        agent_id: int | None = None,
        agent_name: str | None = None,
        mode: str = "agent_autopilot",
        platform: str = "unknown",
        responsible: int | None = None,
        client_id: str | None = None,
        created_by_id: int | None = None,
        timeout: float = 30
    ):
        """
        Create chat for inference.
        TODO: Implement own chat manager for chats, just like I did with catalogs.
        """
        response = await self.router.send(
            "messages:chat:create",
            Serializable[dict[str, Any]]({
                "name": name,
                "agent_id": agent_id,
                "agent_name": agent_name,
                "mode": mode,
                "platform": platform,
                "responsible": responsible,
                "client_id": client_id,
                "created_by_id": created_by_id
            }),
            timeout=timeout,
            expected_response=dict[str, Any]
        )
        
        return response
    
    async def change_chat_agent(
        self,
        chat_id: UUID,
        agent_id: int | None = None,
        agent_name: str | None = None,
        timeout: float = 30
    ) -> dict[str, Any]:
        """
        Change the agent associated with a chat.

        Parameters
        ----------
        chat_id : UUID
            The ID of the chat to change the agent for.
        agent_id : int | None, optional
            The ID of the new agent to associate with the chat, by default None.
        agent_name : str | None, optional
            The name of the new agent to associate with the chat, by default None.
        timeout : float, optional
            The timeout for the request, by default 30.

        Returns
        -------
        dict[str, Any]
            The response from the change agent request.
        """
        response = await self.router.send(
            "messages:chat:change_agent",
            Serializable[dict[str, Any]]({
                "chat_id": str(chat_id),
                "agent_id": agent_id,
                "agent_name": agent_name
            }),
            timeout=timeout,
            expected_response=dict[str, Any]
        )
        
        return response
    
    @overload
    async def invoke_inference(
        self,
        agent_id: int,
        agent_name: str,
        *,
        input_configuration: dict[str, Any] | None = None,
        messages: Sequence[IMessageDTOV2] | None = None,
        stream: Literal[False] = False,
        timeout: float = 200
    ) -> IMessageResponse: ...
    
    @overload
    async def invoke_inference(
        self,
        agent_id: int,
        agent_name: str,
        *,
        input_configuration: dict[str, Any] | None = None,
        messages: Sequence[IMessageDTOV2] | None = None,
        stream: Literal[True] = True,
        timeout: float = 200
    ) -> AsyncIterable[IMessageResponse]: ...
    
    async def invoke_inference(
        self,
        agent_id: int | None = None,
        agent_name: str | None = None,
        *,
        input_configuration: dict[str, Any] | None = None,
        messages: Sequence[IMessageDTOV2] | None = None,
        stream: bool = False,
        timeout: float = 200
    ) -> IMessageResponse | AsyncIterable[IMessageResponse]:
        """
        Invoke inference for a specific agent by its ID or name.
    
        Parameters
        ----------
        agent_id : int | None, optional
            The ID of the agent to invoke inference for, by default None
        agent_name : str | None, optional
            The name of the agent to invoke inference for, by default None
            _description_, by default None
        input_configuration : dict[str, Any] | None, optional
            The configuration for the input, by default None
        messages : Sequence[IMessageDTOV2] | None, optional
            The messages to include in the inference, by default None
        stream : bool, optional
            Whether to stream the response, by default False
        timeout : float, optional
            The timeout for the request, by default 200

        Returns
        -------
        IMessageResponse
            The response from the inference request.

        Raises
        ------
        ValueError
            If neither 'agent_id' nor 'agent_name' is provided.
        ValueError
            If both 'agent_id' and 'agent_name' are provided.
        """
        
        if not agent_id and not agent_name:
            raise ValueError("Required at least one identification specifier, 'agent_id' or 'agent_name'")
        
        if agent_id and agent_name:
            raise ValueError("Cannot find agent by two identification specifiers, use only one!")
        
        if stream:
            iterable_response = await self.router.request_stream("messages:inference:invoke", Serializable[dict[str, Any]]({
                "agent_id": agent_id,
                "agent_name": agent_name,
                "input_configuration": input_configuration,
                "messages": [message.model_dump(mode="json") for message in (messages or [])],
                "stream": stream
            }), timeout=timeout, formatter=lambda x: self.router.convert_message(IMessageResponse, x))
            
            return iterable_response
        
        response = await self.router.send("messages:inference:invoke", Serializable[dict[str, Any]]({
            "agent_id": agent_id,
            "agent_name": agent_name,
            "input_configuration": input_configuration,
            "messages": [message.model_dump(mode="json") for message in (messages or [])],
            "stream": stream
        }), timeout=timeout, expected_response=IMessageResponse)
        
        return response
    
    @overload
    async def invoke_chat_inference(
        self,
        messages: Sequence[IMessageDTOV2],
        chat_id: UUID,
        stream: Literal[False] = False,
        timeout: float = 200
    ) -> IMessageResponse: ...
    
    @overload
    async def invoke_chat_inference(
        self,
        messages: Sequence[IMessageDTOV2],
        chat_id: UUID,
        stream: Literal[True] = True,
        timeout: float = 200
    ) -> AsyncIterable[IMessageResponse]: ...
    
    async def invoke_chat_inference(
        self, 
        messages: Sequence[IMessageDTOV2], 
        chat_id: UUID,
        stream: bool = False,
        timeout: float = 200,
    ) -> IMessageResponse | AsyncIterable[IMessageDTOV2]:
        """
        Invoke inference for a specific chat by its chat_id.

        Parameters
        ----------
        messages : Sequence[IMessageDTOV2]
            The messages to include in the inference.
        chat_id : UUID
            The ID of the chat to invoke inference for.
        stream : bool, optional
            Whether to stream the response, by default False.
        timeout : float, optional
            The timeout for the request, by default 200.

        Returns
        -------
        IMessageResponse
            The response from the inference request.
        """
        for message in messages:
            await self.router.send("messages:append", message, timeout=5)
        
        if stream:
            iterable_response = await self.router.request_stream(
                "messages:chat:invoke", 
                Serializable[dict[str, Any]]({
                    "chat_id": str(chat_id),
                    "stream": stream
                }),
                timeout=timeout,
                formatter=lambda x: self.router.convert_message(IMessageDTOV2, x) if x.payload else None
            )
            
            return iterable_response
        
        response = await self.router.send(
            "messages:chat:invoke", 
            Serializable[dict[str, Any]]({
                "chat_id": str(chat_id),
                "stream": stream
            }),
            timeout=timeout,
            expected_response=IMessageResponse
        )
        
        return response

    async def append_message(self, message: IMessageDTOV2 | Sequence[IMessageDTOV2]) -> None:
        """
        Append a message or a sequence of messages to the current chat.
        
        Parameters
        ----------
        message : IMessageDTOV2 | Sequence[IMessageDTOV2]
            The message or messages to append.
        """
        if isinstance(message, Sequence):
            for msg in message:
                await self.router.emit("messages:append", msg)
        else:
            await self.router.emit("messages:append", message)